package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.io.*;
import java.util.*;
import java.awt.*;            // Only for FileDialog

// **********************************************************************
// **********************************************************************
// class VMPFile encapsulates all the information in a VMP file, and
//         has methods for reading and writing VMP Files.

class VMPFile {
    public static final boolean debug=false;
    
    // Data acessable to the rest of the rknop.jvmap package

    String dir;              // Directory where files are found
    String filename;
    String title;
    int vmpnum;
    MapInfo[] maps;
    TilesInfo tiles; 
    PlayerInfo[] players;
    UnitInfo[] units;
    LineInfo[] lines;
    RulerInfo[] rulers;
    String[] other;

    String errormsg;         // Error message if there was an error

    // **********************************************************************
    // constructor clears out the file

    public VMPFile() {
        clear();
    }

    // **********************************************************************
    // 

    public void clear() {
        dir=new String("");
        filename=new String("");
        title=new String("");
        vmpnum=0;
        maps=new MapInfo[0];
        tiles=new TilesInfo();
        players=new PlayerInfo[0];
        units=new UnitInfo[0];
        lines=new LineInfo[0];
        rulers=new RulerInfo[0];
        other=new String[0];

        errormsg=new String("");
    }

    // **********************************************************************
    // errorMsg() -- returns the error message if there was one,
    //               and then clears it;

    public String errorMsg() {
        String s=new String(errormsg);
        errormsg=new String("");
        return s;
    }

    // **********************************************************************
    // loadVMPFile() prompts for a filename and calls readVMPfile

    public boolean loadVMPFile(Frame frame) {
        if (debug) {
            System.err.println("About to make FileDialog.");
        }
        FileDialog filebox=new FileDialog(frame,"Read VMPFile...",
                                          FileDialog.LOAD);
        if (debug) {
            System.err.println("...about to show FileDialog.");
        }
        if (dir!=null && dir.length()>0) {
            filebox.setDirectory(dir);
        }
        filebox.setVisible(true);
        String loaddir=filebox.getDirectory();
        String loadfile=filebox.getFile();
        if (loadfile==null) return false;
        File vmpfile=new File(loaddir,loadfile);
        boolean success=readVMPFile(vmpfile.getPath());
        if (success) {
            dir=new String(loaddir);
            filename=new String(loadfile);
        }
        return success;
    }

    // **********************************************************************
    // readVMPFile(), given a filename, reads the VMP file into
    //   a Vector of String, and then calls parseVMPFile on that
    //   Vector.

    public boolean readVMPFile(String vmpfilename) {
        FileVecReader vr=null;
        Vector vmpstrings=null;

        try {
            vr=new FileVecReader(vmpfilename);
        } catch (FileNotFoundException ex) {
            errormsg=new String("File not found: "+vmpfilename);
            return false;
        }

        try {
            vmpstrings=vr.readVec();
        } catch (FileNotFoundException ex) {
            errormsg=new String("Error reading "+vmpfilename);
            return false;
        } catch (IOException ex) {
            errormsg=new String("Error reading "+vmpfilename);
            return false;
        }

        vr=null;

        if (parseVMPFile(vmpstrings)) {
            File f=new File(vmpfilename);
            dir=new String(f.getParent());
            filename=new String(f.getName());
            return true;
        }
        else {
            dir="";
            filename="";
            return false;
        }
    }


    // **********************************************************************
    // parseVMPFile() takes a vector of strings that represent a VMP
    //   file and parses them out, creating lots of subelements and reads
    //   lots of images as necessary.

    public boolean parseVMPFile(Vector vmpstrings) {
        String linewords[];
        String cmd;
        int off,next;
        boolean havetiles=false,invmp=false;;
        float vmpversion=0;
        Vector maps=new Vector(1,3);
        Vector players=new Vector(1,3);
        Vector units=new Vector(10,10);
        Vector other=new Vector(10,10);
        
        // Set a few defaults up
        
        clear();


        // Start parsing

        Enumeration vmpenum=vmpstrings.elements();

        // Parse through and find the VERSION string

        invmp=false;
        while (vmpenum.hasMoreElements()) {
            String line=(String)vmpenum.nextElement();

            if (line.equals("BEGIN VMP FILE")) {
                if (invmp) {
                    errormsg=new String("Error: BEGIN VMP FILE appears twice "+
                                        "before END VMP FILE");
                    return false;
                }
                invmp=true;
            }

            else if (line.equals("END VMP FILE")) break;

            else if (line.length()>=7 && 
                     line.substring(0,7).equals("VERSION")) {
                vmpversion=Float.valueOf(line.substring(8)).floatValue();
                if (vmpversion>=2.0 && !invmp) {
                    vmpversion=0;
                    continue;
                }
                if ( ! ( (vmpversion>=1.5 && vmpversion<=1.7) ||
                         (vmpversion==2.0) ) ) {
                    errormsg=new String("Unsupported VMP version "+
                                        String.valueOf(vmpversion)+
                                        System.getProperty("line.separator")+
                                        "Only versions 1.5-1.7, 2.0 "+
                                        "are supported.");
                    return false;
                }
                if (vmpversion>=1.6 && vmpversion<=1.7) {
                    errormsg=new String("Warning: Version 1.6-1.7 VMP files "+
                                        "only partially supported"+
                                        System.getProperty("line.separator")+
                                        "\"LINE\" and some other tags are "+
                                        "not supported.");
                }
                break;
            }
        }
        if (vmpversion<1.5) {
            errormsg=new String("Couldn't figure out VMP version");
            return false;
        }

        if (debug) {
            System.err.println("VMP version = "+String.valueOf(vmpversion));
        }


        // Parse the file from the beginning again

        vmpenum=vmpstrings.elements();

        // If we're a 2.0 or higher file, find the beginning

        if (vmpversion>=2.0) {
            invmp=false;
            while (vmpenum.hasMoreElements()) {
                String line=(String)vmpenum.nextElement();
                
                if (line.equals("BEGIN VMP FILE")) {
                    invmp=true;
                    break;
                }
            }
            if (!invmp) {
                errormsg=new String("Couldn't find BEGIN VMP FILE");
            }
        }


        while (vmpenum.hasMoreElements()) {
            String line=(String)vmpenum.nextElement();

            if (line.equals("END VMP FILE")) break;     // Skip the rest

            if (line.startsWith("#")) continue;          // Comments
        
            linewords=Steege.splitSpaces(line);         // Split at spaces

            if (linewords.length==0) continue;          // Empty line
            cmd=linewords[0];

            if (cmd.equals("VERSION")) {
                // Version has already been parsed;
            }

            else if (cmd.equals("VMPNUM")) {
                vmpnum=Integer.parseInt(linewords[1]);
            }

            else if (cmd.equals("TITLE")) {
                title=new String(line.substring(6));
            }

            else if (cmd.equals("MAP")) {
                if (vmpversion>=2.0) {
                    linewords=Steege.splitSpaces(line,8);
                    if (linewords.length<7) {
                        errormsg=new String("Error parsming MAP");
                        return false;
                    }
                    MapInfo map=new MapInfo();
                    map.number=Integer.parseInt(linewords[1]);
                    map.x0=Integer.parseInt(linewords[2]);
                    map.y0=Integer.parseInt(linewords[3]);
                    map.width=Integer.parseInt(linewords[4]);
                    map.height=Integer.parseInt(linewords[5]);
                    map.filename=new String(linewords[6]);
                    if (linewords.length>7) {
                        map.title=linewords[7];
                    } else {
                        map.title=new String("");
                    }
                    maps.addElement(map);
                }
                else {
                    if (linewords.length != 5) {
                        errormsg=new String("Error parsing MAP");
                        return false;
                    }
                    if (maps.size()>0) {
                        errormsg=new String("Error: >1 Map for VMP<2.0");
                        return false;
                    }
                    MapInfo map=new MapInfo();
                    map.number=0;
                    map.x0=0;
                    map.y0=0;
                    map.width=0;
                    map.height=0;
                    map.filename=new String(linewords[4]);
                    map.title=new String(linewords[4]);
                    maps.addElement(map);
                }
            }

            else if (cmd.equals("TILES")) {
                if (linewords.length != 4) {
                    errormsg=new String("Error parsing TILES.");
                    return false;
                }
                tiles.rows=Integer.parseInt(linewords[1]);
                tiles.columns=Integer.parseInt(linewords[2]);
                tiles.filename=new String(linewords[3]);
                havetiles=true;
            }

            else if (cmd.equals("PLAYER")) {
                linewords=Steege.splitSpaces(line,5);
                if (linewords.length<5) {
                    errormsg=new String("Error parsing PLAYER");
                    return false;
                }
                PlayerInfo player=new PlayerInfo();
                player.number=Integer.parseInt(linewords[1]);
                player.email=linewords[2];
                player.password=linewords[3];
                player.name=linewords[4];
                players.addElement(player);
            }

            else if (linewords[0].equals("UNIT")) {
                if (vmpversion>=2.0) {
                    linewords=Steege.splitSpaces(line,9);
                    if (linewords.length<8) {
                        errormsg=new String("Error parsing UNIT");
                        return false;
                    }
                    UnitInfo unit=new UnitInfo();
                    unit.ordinal=Integer.parseInt(linewords[1]);
                    unit.top=Integer.parseInt(linewords[2]);
                    unit.left=Integer.parseInt(linewords[3]);
                    unit.map=Integer.parseInt(linewords[4]);
                    unit.index=Integer.parseInt(linewords[5]);
                    unit.owner=Integer.parseInt(linewords[6]);
                    unit.facing=Float.valueOf(linewords[7]).floatValue();
                    if (linewords.length>8) {
                        unit.comment=linewords[8];
                    } else {
                        unit.comment=new String("");
                    }
                    units.addElement(unit);
                }
                else {
                    linewords=Steege.splitSpaces(line,8);
                    if (linewords.length<5) {
                        errormsg=new String("Error parsing UNIT");
                        return false;
                    }
                    UnitInfo unit=new UnitInfo();
                    unit.ordinal=units.size();
                    unit.top=Integer.parseInt(linewords[1]);
                    unit.left=Integer.parseInt(linewords[2]);
                    unit.map=0;
                    unit.index=Integer.parseInt(linewords[4]);
                    if (linewords.length>5) {
                        unit.owner=Integer.parseInt(linewords[5]);
                    } else {
                        unit.owner=0;
                    }
                    if (linewords.length>7) {
                        unit.comment=linewords[7];
                    } else {
                        unit.comment=new String("");
                    }
                    units.addElement(unit);
                }
            }

            else {
                if (  linewords[0].equals("DICE") || 
                      linewords[0].equals("LINE") ||
                      linewords[0].equals("TABLE") ||
                      linewords[0].equals("RULER") ||
                      linewords[0].equals("SNAP") ||
                      linewords[0].equals("ENCRYPT")  ) {
                    String text=new String(line);
                    other.addElement(text);
                } else {
                    if (debug) {
                        System.err.println("Ignoring line: "+line);
                    }
                }
            }
        }

        // A little bit of terminal error checking

        if (maps.size()<1) {
            errormsg=new String("Error in VMP File, no MAP found.");
            return false;
        }

        if (!havetiles) {
            errormsg=new String("Error in VMP file, no TILES found.");
            return false;
        }


        // Fill our instance variable arrays from the vectors we've read

        this.maps=new MapInfo[maps.size()];
        maps.copyInto(this.maps);
        this.players=new PlayerInfo[players.size()];
        players.copyInto(this.players);
        this.units=new UnitInfo[units.size()];
        units.copyInto(this.units);
        this.other=new String[other.size()];
        other.copyInto(this.other);

        return true;
    }

    // **********************************************************************
    // writeVMP15() -- write a VMP version 1.5 file
    //    It assumes that it's safe to overwrite the file if it's there,
    //    and doesn't even check.

    public void writeVMP15(String filename) throws Exception {
        String line,yword,xword,inplayword,indexword,ownerword,reservedword;
        int i;
        
        if (maps.length!=1) {
            throw new Exception("writeVMP15 Error: "+
                                "Must have exactly one map.");
        }

        File ofile=new File(filename);

        File f=new File(ofile.getAbsolutePath());
        dir=f.getParent();
        this.filename=ofile.getName();

        FileWriter fw=new FileWriter(ofile);
        PrintWriter pw=new PrintWriter(fw,true);

        if (debug) {
            System.err.println("Starting to spew text.");
        }

        pw.print("# V_MAP save file written by JV_Map "+JV_Map.jvmapversion);
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("VERSION 1.5");
        pw.print((char)0x0d);
        pw.print((char)0x0a);

        // Print the MAP line

        pw.print("# MAP step_size max_top max_left name");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        line=new String("MAP 0 0 0 "+maps[0].filename);
        pw.print(line);
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        
        // Print the TILES line
      
        pw.print("#");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("# TILES rows columns name");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        line=new String("TILES "+String.valueOf(tiles.rows)+" "+
                        String.valueOf(tiles.columns)+" "+
                        tiles.filename);
        pw.print(line);
        pw.print((char)0x0d);
        pw.print((char)0x0a);
      
        // Stacking...

        pw.print("#");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("# STACKING has never been understood by JV_Map :)");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("STACKING ON");
        pw.print((char)0x0d);
        pw.print((char)0x0a);

        // Other, miscellaneous unparsed lines get skipped for
        // VMP 1.5 files

        // Finally, print out all the units

        if (debug) {
            System.err.println("Spewing units.");
        }

        pw.print("#");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("# UNIT top left in_play bitmap_index owner "+
                 "reserved comment");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        for (i=0 ; i<units.length ; ++i) {

            yword=new String(String.valueOf(units[i].top));
            while (yword.length()<7) yword=yword+" ";
            xword=new String(String.valueOf(units[i].left));
            while (xword.length()<7) xword=xword+" ";
            inplayword=new String(String.valueOf(1));
            while (inplayword.length()<5) inplayword=inplayword+" ";
            indexword=new String(String.valueOf(units[i].index));
            while (indexword.length()<6) indexword=indexword+" ";
            ownerword=new String(String.valueOf(units[i].owner));
            while (ownerword.length()<5) ownerword=ownerword+" ";
            reservedword=new String(String.valueOf(0));
            while (reservedword.length()<5)
                reservedword=reservedword+" ";
            
            if (units[i].comment!=null) {
                line=new String("UNIT "+yword+xword+inplayword+
                                indexword+ownerword+reservedword+
                                units[i].comment);
            } else {
                line=new String("UNIT "+yword+xword+inplayword+
                                indexword+ownerword+reservedword);
            }
            pw.print(line);
            pw.print((char)0x0d);
            pw.print((char)0x0a);
        }

        if (debug) {
            System.err.println("Done spewing text.");
        }

        fw.close();
    }

    // **********************************************************************
    // writeVMP20() -- write a VMP version 2.0 file
    //    It assumes that it's safe to overwrite the file if it's there,
    //    and doesn't even check.

    public void writeVMP20(String filename) throws Exception {
        int i;
        String line;
        
        File ofile=new File(filename);

        File f=new File(ofile.getAbsolutePath());
        dir=f.getParent();
        this.filename=ofile.getName();

        FileWriter fw=new FileWriter(ofile);
        PrintWriter pw=new PrintWriter(fw,true);

        if (debug) {
            System.err.println("Starting to spew text.");
        }

        pw.print("BEGIN VMP FILE");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("# VMP 2.0 file written by JV_Map "+JV_Map.jvmapversion);
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("VERSION 2.0");
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("VMPNUM "+String.valueOf(vmpnum));
        pw.print((char)0x0d);
        pw.print((char)0x0a);
        pw.print("TITLE "+title);
        pw.print((char)0x0d);
        pw.print((char)0x0a);

        // Maps

        for (i=0 ; i<maps.length ; ++i) {
            line="MAP "+String.valueOf(maps[i].number)+" ";
            line+=String.valueOf(maps[i].x0)+" ";
            line+=String.valueOf(maps[i].y0)+" ";
            line+=String.valueOf(maps[i].width)+" ";
            line+=String.valueOf(maps[i].height)+" ";
            line+=maps[i].filename;
            if (maps[i].title!=null && maps[i].title.length()>0) {
                line+=" "+maps[i].title;
            }
            pw.print(line);
            pw.print((char)0x0d);
            pw.print((char)0x0a);
        }

        // Tiles set

        line="TILES "+String.valueOf(tiles.rows)+" ";
        line+=String.valueOf(tiles.columns)+" ";
        line+=tiles.filename;
        pw.print(line);
        pw.print((char)0x0d);
        pw.print((char)0x0a);

        // Players

        for (i=0 ; i<players.length ; ++i) {
            line="PLAYER "+String.valueOf(players[i].number)+" ";
            line+=players[i].email+" ";
            line+=players[i].password+" ";
            line+=players[i].name;
            pw.print(line);
            pw.print((char)0x0d);
            pw.print((char)0x0a);
        }

        // Miscellaneous unparsed lines, which we preserve just in case
        // Eventually, I'd like to fully implement the spec so that 
        // there won't need to be any of these

        for (i=0 ; i<other.length ; ++i) {
            pw.print(other[i]);
            pw.print((char)0x0d);
            pw.print((char)0x0a);
        }

        // Units

        for (i=0 ; i<units.length ; ++i) {
            line="UNIT "+String.valueOf(units[i].ordinal)+" ";
            line+=String.valueOf(units[i].top)+" ";
            line+=String.valueOf(units[i].left)+" ";
            line+=String.valueOf(units[i].map)+" ";
            line+=String.valueOf(units[i].index)+" ";
            line+=String.valueOf(units[i].owner)+" ";
            line+=String.valueOf(units[i].facing);
            if (units[i].comment!=null && units[i].comment.length()>0) {
                line+=" "+units[i].comment;
            }
            pw.print(line);
            pw.print((char)0x0d);
            pw.print((char)0x0a);
        }

        // Lines, ruler, dice, tables

        // End
        
        pw.print("END VMP FILE");
        pw.print((char)0x0d);
        pw.print((char)0x0a);

        fw.close();
    }

}    

